#include "stdafx.h"
/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc

  Authors: Sebastian Deorowicz and Agnieszka Debudaj-Grabysz

  Version: 0.3
  Date   : 2012-11-29
*/

#include "kb_compactor.h"
#include <algorithm>
#include <numeric>
#include <tuple>

using namespace std;

//----------------------------------------------------------------------------------
// Constructor
CKmerPackageCompactor::CKmerPackageCompactor(CKMCParams &Params, CKMCQueues &Queues)
{
	q_package          = Queues.q_package;
	m_disk             = Queues.dm;
	bd				   = Queues.bd;
	write_buffer_queue = Queues.write_buffer_queue;
	pmm_bins           = Queues.pmm_bins;
	pmm_cnts_small     = Queues.pmm_cnts_small;
	pmm_cnts_large     = Queues.pmm_cnts_large;

	mm_write_buffer    = Queues.mm_write_buffer;

	n_disks			   = Params.n_disks;

	buffer = NULL;
}

//----------------------------------------------------------------------------------
// Destructor
CKmerPackageCompactor::~CKmerPackageCompactor()
{
}

//----------------------------------------------------------------------------------
// Process the queue of packages
void CKmerPackageCompactor::ProcessQueue()
{
	CBinPackageQueue::parts_t *package;

	// Process the queue
	while(!q_package->completed())
	{
		int32 bin_id;
		int32 disk_id;
		FILE *file = NULL;

		if(q_package->pop(bin_id, disk_id, file, package))
		{
			Compact(bin_id, package);
			write_buffer_queue[disk_id]->push(file, buffer, buffer_size);
		}
	}
	for(int i = 0; i < n_disks; ++i)
		write_buffer_queue[i]->mark_completed();
}

//----------------------------------------------------------------------------------
// Read counters for prefixes
void CKmerPackageCompactor::ReadPrefCounters(uint32 pref_size, uchar *buf, uint32 *cnts, uint32 &part_records)
{
	uchar *t = buf + 4;
	
	part_records = *t++;
	part_records = (part_records << 8) + *t++;
	part_records = (part_records << 8) + *t++;
	part_records = (part_records << 8) + *t++;

	// Read counters
	uint32 cnt;
	for(uint32 i = 0; i < pref_size; ++i)
	{
		cnt = *t++;
		cnt = (cnt << 8) + *t++;

		cnts[i]  = cnt;			
	}
}

//----------------------------------------------------------------------------------
// Compact bin parts
void CKmerPackageCompactor::Compact(int32 bin_id, CBinPackageQueue::parts_t *package)
{
	uint32 count_len = 0;
	uint32 rec_len = 0;
	uint32 quality_len = 0;
	uint32 n_records = 0;
	uint32 *pref_pack_cnts, *cum_pref_pack_cnts;
	uint32 *new_pref_pack_cnts;

	count_len = *get<0>(package->front());
	uint32 pref_size = 1 << (count_len * 2);
	uint32 new_pref_size = pref_size * 256;

	pmm_cnts_small->reserve(pref_pack_cnts);
	pmm_cnts_small->reserve(cum_pref_pack_cnts);
	pmm_cnts_large->reserve(new_pref_pack_cnts);

	fill_n(cum_pref_pack_cnts, pref_size, 0);
	fill_n(new_pref_pack_cnts, new_pref_size, 0);
	buffer_size = 0;

	// Count statistics of k-mers
	for(auto p = package->begin(); p != package->end(); ++p)
	{
		uchar *t = get<0>(*p);
		uint32 part_records;

		count_len   = *t++;
		quality_len = *t++;
		quality_len >>= 4;
		rec_len     = *t++;
		rec_len     = (rec_len << 8) + *t++;

		ReadPrefCounters(pref_size, get<0>(*p), pref_pack_cnts, part_records);
		for(uint32 i = 0; i < pref_size; ++i)
			cum_pref_pack_cnts[i] += pref_pack_cnts[i];
		partial_sum(pref_pack_cnts, pref_pack_cnts+pref_size, pref_pack_cnts);
		t += 4 + 2 * pref_size;

		n_records   += part_records;
		buffer_size += get<1>(*p);

		// Read data for new counters
		uint32 byte;

		uint32 pref_id  = 0;
		uint32 skip_len = rec_len + quality_len;
		for(uint32 i = 0; i < part_records; ++i)
		{
			byte = *t;
			t += skip_len;

			while(i >= pref_pack_cnts[pref_id])
				pref_id++;
			new_pref_pack_cnts[(pref_id << 8) + byte]++;
		}
	}

	uint32 max_new_bucket_size = *max_element(new_pref_pack_cnts, new_pref_pack_cnts+new_pref_size);
	uint32 new_cnts_field_size = (max_new_bucket_size < (1ul << 8)) ? 1 : (max_new_bucket_size < (1ul << 16)) ? 2 : (max_new_bucket_size < (1ul << 24)) ? 3 : 4;

	uint32 max_cum_bucket_size = *max_element(cum_pref_pack_cnts, cum_pref_pack_cnts+pref_size);
	uint32 cum_cnts_field_size = (max_cum_bucket_size < (1ul << 8)) ? 1 : (max_cum_bucket_size < (1ul << 16)) ? 2 : (max_cum_bucket_size < (1ul << 24)) ? 3 : 4;

	// Check whether it is better to remove one more symbol
	uint64 act_size = (n_records * (rec_len + quality_len)) + package->size() * (8 + 2 * pref_size);
	uint64 new_size = (n_records * (rec_len + quality_len - 1)) + 8 + new_cnts_field_size * new_pref_size;
	uint64 cum_size = (n_records * (rec_len + quality_len)) + 8 + cum_cnts_field_size * pref_size;

	if(cum_size <= new_size)			// merge package without compressing (only reduces the counter array)
	{
		buffer_size = cum_size;
		mm_write_buffer->force_increase(buffer_size);
		buffer = new uchar[buffer_size];

		uchar *ptr = buffer;

		// Store description
		*ptr++ = (uchar) (count_len);
		*ptr++ = (uchar) ((quality_len << 4) + cum_cnts_field_size);
		*ptr++ = (uchar) (rec_len >> 8);
		*ptr++ = (uchar) (rec_len & 0xFF);

		*ptr++ = (uchar) (n_records >> 24);
		*ptr++ = (uchar) ((n_records >> 16) & 0xFF);
		*ptr++ = (uchar) ((n_records >> 8) & 0xFF);
		*ptr++ = (uchar) (n_records & 0xFF);

		// Store counters for prefixes
		if(cum_cnts_field_size == 1)
			for(uint32 i = 0; i < pref_size; ++i)
				*ptr++ = cum_pref_pack_cnts[i] & 0xFF;
		else if(cum_cnts_field_size == 2)
			for(uint32 i = 0; i < pref_size; ++i)
			{
				*ptr++ = (cum_pref_pack_cnts[i] >> 8) & 0xFF;
				*ptr++ = (cum_pref_pack_cnts[i]     ) & 0xFF;
			}
		else if(cum_cnts_field_size == 3)
			for(uint32 i = 0; i < pref_size; ++i)
			{
				*ptr++ = (cum_pref_pack_cnts[i] >> 16) & 0xFF;
				*ptr++ = (cum_pref_pack_cnts[i] >>  8) & 0xFF;
				*ptr++ = (cum_pref_pack_cnts[i]      ) & 0xFF;
			}
		else if(cum_cnts_field_size == 4)
			for(uint32 i = 0; i < pref_size; ++i)
			{
				*ptr++ = (cum_pref_pack_cnts[i] >> 24) & 0xFF;
				*ptr++ = (cum_pref_pack_cnts[i] >> 16) & 0xFF;
				*ptr++ = (cum_pref_pack_cnts[i] >>  8) & 0xFF;
				*ptr++ = (cum_pref_pack_cnts[i]      ) & 0xFF;
			}

		// Store suffixes
		partial_sum(cum_pref_pack_cnts, cum_pref_pack_cnts+pref_size, cum_pref_pack_cnts);
		for(uint32 i = pref_size-1; i > 0; --i)
			cum_pref_pack_cnts[i] = cum_pref_pack_cnts[i-1];
		cum_pref_pack_cnts[0] = 0;
		
		uint32 copy_len = rec_len + quality_len;
		uint32 skip_len = rec_len + quality_len;
		int pck_id = -1;
		for(auto p = package->begin(); p != package->end(); p = package->begin())
		{
			pck_id++;

			uint32 part_records;
			ReadPrefCounters(pref_size, get<0>(*p), pref_pack_cnts, part_records);
			partial_sum(pref_pack_cnts, pref_pack_cnts+pref_size, pref_pack_cnts);
			uchar *t = get<0>(*p) + 8 + 2 * pref_size;
			
			// Distribute data
			uint32 pref_id = 0;
			uint32 act_pref;
			for(uint32 i = 0; i < part_records; ++i)
			{
				while(i >= pref_pack_cnts[pref_id])
					pref_id++;

				act_pref = pref_id;

				copy(t, t+copy_len, ptr+cum_pref_pack_cnts[act_pref]*copy_len);
				cum_pref_pack_cnts[act_pref]++;

				t += skip_len;
			}
			pmm_bins->free(get<0>(*p));
			package->pop_front();
		}

		// Actualize the information about sizes of bins
		bd->insert(bin_id, NULL, "", 0, (int64) cum_size - (int64) act_size, 0);
	}
	else							// recompress
	{
		buffer_size = new_size;
		mm_write_buffer->force_increase(buffer_size);
		buffer = new uchar[buffer_size];

		rec_len -= 1;

		uchar *ptr = buffer;

		// Store description
		*ptr++ = (uchar) (count_len + 4);
		*ptr++ = (uchar) ((quality_len << 4) + new_cnts_field_size);
		*ptr++ = (uchar) (rec_len >> 8);
		*ptr++ = (uchar) (rec_len & 0xFF);

		*ptr++ = (uchar) (n_records >> 24);
		*ptr++ = (uchar) ((n_records >> 16) & 0xFF);
		*ptr++ = (uchar) ((n_records >> 8) & 0xFF);
		*ptr++ = (uchar) (n_records & 0xFF);

		// Store counters for prefixes
		if(new_cnts_field_size == 1)
			for(uint32 i = 0; i < new_pref_size; ++i)
				*ptr++ = new_pref_pack_cnts[i] & 0xFF;
		else if(new_cnts_field_size == 2)
			for(uint32 i = 0; i < new_pref_size; ++i)
			{
				*ptr++ = (new_pref_pack_cnts[i] >> 8) & 0xFF;
				*ptr++ = (new_pref_pack_cnts[i]     ) & 0xFF;
			}
		else if(new_cnts_field_size == 3)
			for(uint32 i = 0; i < new_pref_size; ++i)
			{
				*ptr++ = (new_pref_pack_cnts[i] >> 16) & 0xFF;
				*ptr++ = (new_pref_pack_cnts[i] >>  8) & 0xFF;
				*ptr++ = (new_pref_pack_cnts[i]      ) & 0xFF;
			}
		else if(new_cnts_field_size == 4)
			for(uint32 i = 0; i < new_pref_size; ++i)
			{
				*ptr++ = (new_pref_pack_cnts[i] >> 24) & 0xFF;
				*ptr++ = (new_pref_pack_cnts[i] >> 16) & 0xFF;
				*ptr++ = (new_pref_pack_cnts[i] >>  8) & 0xFF;
				*ptr++ = (new_pref_pack_cnts[i]      ) & 0xFF;
			}

		// Store suffixes
		partial_sum(new_pref_pack_cnts, new_pref_pack_cnts+new_pref_size, new_pref_pack_cnts);
		for(uint32 i = new_pref_size-1; i > 0; --i)
			new_pref_pack_cnts[i] = new_pref_pack_cnts[i-1];
		new_pref_pack_cnts[0] = 0;
		
		uint32 copy_len = rec_len + quality_len;
		uint32 skip_len = rec_len + 1 + quality_len;
		int pck_id = -1;
		for(auto p = package->begin(); p != package->end(); p = package->begin())
		{
			pck_id++;

			uint32 part_records;
			ReadPrefCounters(pref_size, get<0>(*p), pref_pack_cnts, part_records);
			partial_sum(pref_pack_cnts, pref_pack_cnts+pref_size, pref_pack_cnts);
			uchar *t = get<0>(*p) + 8 + 2 * pref_size;
			
			// Read data for new counters
			uint32 byte;
			uint32 pref_id = 0;
			uint32 act_pref;
			for(uint32 i = 0; i < part_records; ++i)
			{
				byte = *t;

				while(i >= pref_pack_cnts[pref_id])
					pref_id++;

				act_pref = (pref_id << 8) + byte;

				copy(t+1, t+1+copy_len, ptr+new_pref_pack_cnts[act_pref]*copy_len);
				new_pref_pack_cnts[act_pref]++;

				t += skip_len;
			}
			pmm_bins->free(get<0>(*p));
			package->pop_front();
		}

		// Actualize the information about sizes of bins
		bd->insert(bin_id, NULL, "", 0, (int64) new_size - (int64) act_size, 0);
	}

	pmm_cnts_small->free(pref_pack_cnts);
	pmm_cnts_small->free(cum_pref_pack_cnts);
	pmm_cnts_large->free(new_pref_pack_cnts);

	delete package;
}


//************************************************************************************************************
// CWKmerPackageCompactor - wrapper
//************************************************************************************************************

//----------------------------------------------------------------------------------
// Constructor
CWKmerPackageCompactor::CWKmerPackageCompactor(CKMCParams &Params, CKMCQueues &Queues)
{
	kpc = new CKmerPackageCompactor(Params, Queues);
}

//----------------------------------------------------------------------------------
// Destructor
CWKmerPackageCompactor::~CWKmerPackageCompactor()
{
	delete kpc;
}

//----------------------------------------------------------------------------------
// Execution
void CWKmerPackageCompactor::operator()()
{
	kpc->ProcessQueue();
}
