#include "stdafx.h"
/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc
  
  Authors: Sebastian Deorowicz and Agnieszka Debudaj-Grabysz
  
  Version: 0.3
  Date   : 2012-11-29
*/
#include <algorithm>
#include <numeric>
#include <iostream>
#include <boost/lexical_cast.hpp>
#include "kb_storer.h"

using namespace std;

extern uint64 total_reads;

//************************************************************************************************************
// CKmerBinStorer - storer for bins
//************************************************************************************************************

//----------------------------------------------------------------------------------
// Constructor
CKmerBinStorer::CKmerBinStorer(CKMCParams &Params, CKMCQueues &Queues)
{
	mm					= Queues.mm;
	n_bins			    = Params.n_bins;
	q_part			    = Queues.bpq;
	q_package           = Queues.q_package;
	bd                  = Queues.bd;
	working_directories = Params.working_directories;

	files                  = NULL;
	f_disks			       = NULL;
	buf_sizes		       = NULL;
	buffer_size_bytes      = 0;
	max_buf_size		   = 0;
	max_buf_size_id		   = 0;
	max_mem_buffer         = Params.max_mem_storer;
	max_mem_single_package = Params.max_mem_storer_pkg;

	buffer = new elem_t*[n_bins];
	for(int i = 0; i < n_bins; ++i)
		buffer[i] = NULL;
}

//----------------------------------------------------------------------------------
// Destructor
CKmerBinStorer::~CKmerBinStorer()
{
	Release();
}

//----------------------------------------------------------------------------------
// Write ends of bins and release memory
void CKmerBinStorer::Release()
{
	if(!files)
		return;
	for(int i = 0; i < n_bins; ++i)
		if(buffer[i])
			delete buffer[i];

	delete[] buffer;
	buffer = NULL;	

	delete[] files;
	files = NULL;

	delete[] f_disks;
	f_disks = NULL;

	delete[] buf_sizes;
	buf_sizes = NULL;
	
	cout << "\n";
}

//----------------------------------------------------------------------------------
// Put buffer items to the queue
void CKmerBinStorer::ReleaseBuffer()
{
	for(int i = 0; i < n_bins; ++i)
		if(files[i])
			PutBinToCompact(i);

	for(int i = n_bins-1; i >= 0; --i)
		if(buffer[i])
		{
			delete buffer[i];
			buffer[i] = NULL;
		}
}

//----------------------------------------------------------------------------------
// Return name of a file related to a kmer of given id. Also return disk no. for that kmer
string CKmerBinStorer::GetName(int n, uint32 &c_disk)
{
	string s_tmp = boost::lexical_cast<string>(n);
	while(s_tmp.length() < 5)
		s_tmp = string("0") + s_tmp;
	
	// If no. of temporary directories is 1, just return the name of kmer file
	if(working_directories.size() == 1)
	{
		c_disk = 0;
		if(*working_directories[0].rbegin() != '/' && *working_directories[0].rbegin() != '\\')
			working_directories[0] += "/";
		return working_directories[0]  + "kmc_" + s_tmp + ".bin";
	}

	// Compute the disk id of a given bin. 
	// Disks are allocated in a manner: 0, 1, 2, ... max, max, max-1, ..., 1, 0, 0, 1, 2, ...
	uint32 n_disks = (uint32) working_directories.size();

	if(n < 4)
		;
	else if(n < n_bins/4 + n_bins/16)
		n /= 4;
	else
		n /= 16;

	c_disk = n % (2 * n_disks);
	c_disk = MIN(c_disk, 2*n_disks-1-c_disk);

	if(*working_directories[c_disk].rbegin() != '/' && *working_directories[c_disk].rbegin() != '\\')
		working_directories[c_disk] += "/";
	return working_directories[c_disk]  + "kmc_" + s_tmp + ".bin";
}

//----------------------------------------------------------------------------------
// Check wheter it is necessary to store some bin to a HDD
void CKmerBinStorer::CheckBuffer()
{
	int32 i;

	if(buffer_size_bytes < max_mem_buffer && max_buf_size < max_mem_single_package)
		return;

	PutBinToCompact(max_buf_size_id);

	buf_sizes[max_buf_size_id] = 0;

	max_buf_size    = buf_sizes[0];
	max_buf_size_id = 0;
	for(i = 1; i < n_bins; ++i)
		if(buf_sizes[i] > max_buf_size)
		{
			max_buf_size    = buf_sizes[i];
			max_buf_size_id = i;
		}
}

//----------------------------------------------------------------------------------
// Send bin to compaction phase
void CKmerBinStorer::PutBinToCompact(uint32 n)
{
	if(buf_sizes[n])
	{
		q_package->push(n, f_disks[n], files[n], buffer[n], buf_sizes[n]);
		buffer_size_bytes -= buf_sizes[n];
	}

	buffer[n] = NULL;
}

//----------------------------------------------------------------------------------
// Open temporary files for all bins
bool CKmerBinStorer::OpenFiles()
{
	string f_name;
	uint32 c_disk;

	files     = new FILE*[n_bins];
	f_disks   = new uint32[n_bins];
	buf_sizes = new uint64[n_bins];

	for(int i = 0; i < n_bins; ++i)
	{
		f_name = GetName(i, c_disk);
		f_disks[i] = c_disk;
		buf_sizes[i] = 0;

		if((i < 4) || (i < n_bins/4 + n_bins/16 && i % 4 == 0) || (i % 16 == 0))
		{
			files[i] = fopen(f_name.c_str(), "wb+");	
			if(!files[i])
			{
				cout << "Error: Cannot open temporary file " << f_name << "\n";
				exit(1);
			}
			bd->insert(i, files[i], f_name, c_disk, 0, 0);
		}
		else
			files[i] = NULL;
	}

	return true;
}

//----------------------------------------------------------------------------------
// 
void CKmerBinStorer::ProcessQueue()
{
	// Process the queue
	while(!q_part->completed())
	{
		int32 bin_id;
		uchar *part;
		uint32 true_size;
		uint32 alloc_size;

		if(q_part->pop(bin_id, part, true_size, alloc_size))
		{
			if(!buffer[bin_id])
				buffer[bin_id] = new elem_t;
			buffer[bin_id]->push_back(make_tuple(part, true_size, alloc_size));
			buffer_size_bytes += alloc_size;
			buf_sizes[bin_id] += alloc_size;

			if(buf_sizes[bin_id] > max_buf_size)
			{
				max_buf_size    = buf_sizes[bin_id];
				max_buf_size_id = bin_id;
			}

			CheckBuffer();
		}
	}

	// Move all remaining parts to queue
	ReleaseBuffer();

	q_package->mark_completed();
}


//************************************************************************************************************
// CWKmerBinStorer - wrapper
//************************************************************************************************************

//----------------------------------------------------------------------------------
// Constructor
CWKmerBinStorer::CWKmerBinStorer(CKMCParams &Params, CKMCQueues &Queues)
{
	kbs = new CKmerBinStorer(Params, Queues);
	kbs->OpenFiles();
}

//----------------------------------------------------------------------------------
// Destructore
CWKmerBinStorer::~CWKmerBinStorer()
{
	delete kbs;
}

//----------------------------------------------------------------------------------
// Execution
void CWKmerBinStorer::operator()()
{
	kbs->ProcessQueue();
}

// ***** EOF
