/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc

  Authors: Sebastian Deorowicz and Agnieszka Debudaj-Grabysz

  Version: 0.3
  Date   : 2012-11-29
*/

#ifndef _KMC_H
#define _KMC_H

#include "defs.h"
#include "params.h"
#include "kmer.h"
#include <iostream>
#include <iomanip>
#include <string>
#include <vector>
#include <numeric>
#include "queues.h"
#include "timer.h"
#include "fastq_reader.h"
#include "kb_collector.h"
#include "kb_completer.h"
#include "kb_reader.h"
#include "kb_sorter.h"
#include "kb_storer.h"
#include "kb_compactor.h"
#include "kb_writer.h"
#include "splitter.h"
#include "libs/asmlib.h"
#include <boost/filesystem.hpp>

using namespace std;


template <typename KMER_T, unsigned SIZE> class CKMC {
	bool initialized;

	CStopWatch w1, w2;

	// Parameters (input and internal)
	CKMCParams Params;

	// Memory monitor and queues
	CKMCQueues Queues;

	// Thread groups
	vector<thread> gr1_1, gr1_2, gr1_3, gr1_4, gr1_5;		// thread groups for 1st stage
	vector<thread> gr2_1, gr2_2, gr2_3;						// thread groups for 2nd stage

	uint64 n_unique, n_cutoff_min, n_cutoff_max, n_total, n_reads, tmp_size;

	// Threads
	vector<CWFastqReader*> w_fastqs;
	vector<CWSplitter<KMER_T, SIZE>*> w_splitters;
	CWKmerBinStorer *w_storer;
	vector<CWKmerPackageCompactor*> w_compactors;
	vector<CWKmerBinWriter*> w_writers;

	vector<CWKmerBinReader<KMER_T, SIZE>*> w_readers;
	vector<CWKmerBinSorter<KMER_T, SIZE>*> w_sorters;
	CWKmerBinCompleter *w_completer;

	void AdjustToHardware(int32 cores);
	bool AdjustMemoryLimits();

	void ShowSettings();

public:
	CKMC();
	~CKMC();
	
	void SetParams(CKMCParams &_Params);
	bool Process();
	void GetStats(double &time1, double &time2, uint64 &_n_unique, uint64 &_n_cutoff_min, uint64 &_n_cutoff_max, uint64 &_n_total, uint64 &_n_reads, uint64 &_tmp_size);
};


//----------------------------------------------------------------------------------
template <typename KMER_T, unsigned SIZE> CKMC<KMER_T, SIZE>::CKMC()
{
// OpenMP support is a must, so do not compile if it is not supported
#if !defined(_OPENMP)
	BOOST_STATIC_ASSERT_MSG(false, "You need to use OpenMP");
#endif

	initialized   = false;
	Params.kmer_len      = 0;
	Params.n_readers     = 1;
	Params.n_splitters   = 1;
	Params.n_sorters     = 1;
	Params.n_omp_threads = 1;
	Params.n_compactors  = 1;
}

//----------------------------------------------------------------------------------
template <typename KMER_T, unsigned SIZE> CKMC<KMER_T, SIZE>::~CKMC()
{

}

//----------------------------------------------------------------------------------
// Set params of the k-mer counter
template <typename KMER_T, unsigned SIZE> void CKMC<KMER_T, SIZE>::SetParams(CKMCParams &_Params)
{
	Params = _Params;

	Params.kmer_len	= Params.p_k;
	Params.verbose	= Params.p_verbose;

	// Technical parameters related to temporary files
	if(Params.p_p1 == 3)
	{
		Params.prefix_len        = 3;
		Params.bin_part_size     = 1 << 15;
	}
	else if(Params.p_p1 == 5)
	{
		Params.prefix_len        = 5;
		Params.bin_part_size     = 1 << 14;
	}
	else			// default p1=4
	{
		Params.prefix_len        = 4;
		Params.bin_part_size     = 1 << 15;
	}

	Params.n_bins            = (1 << (2 * Params.prefix_len)) * 16;

	// Thresholds for counters
	Params.cutoff_min   = Params.p_ci;
	Params.cutoff_max   = Params.p_cx;
	Params.counter_max  = Params.p_cs;
	Params.use_quake    = Params.p_quake;

	Params.lowest_quality = Params.p_quality;
	Params.both_strands   = Params.p_both_strands;

	// Technical parameters related to no. of threads and memory usage
	if(Params.p_sf && Params.p_sp && Params.p_so && Params.p_sc && Params.p_sr)
	{
		Params.n_readers     = NORM(Params.p_sf, 1, 32);
		Params.n_splitters   = NORM(Params.p_sp, 1, 32);
		Params.n_compactors  = NORM(Params.p_sc, 1, 32);
		Params.n_sorters     = NORM(Params.p_sr, 1, 32);
		Params.n_omp_threads = NORM(Params.p_so, 1, 32);
	}
	else
	{
		// Adjust the number of threads according to the current hardware
		Params.n_threads = Params.p_t;
		AdjustToHardware(Params.n_threads);
	}

	Params.max_mem_size  = NORM(((uint64) Params.p_m) << 30, 4ull << 30, 1024ull << 30);

	Params.n_disks			= (uint32) Params.working_directories.size();
	Params.is_fasta			= Params.p_fasta;

	Params.KMER_T_size = sizeof(KMER_T);

	initialized = true; 

	SetMemcpyCacheLimit(8);			// Sets the asmlib's memcpy function to make copy without use of cache memory
}

//----------------------------------------------------------------------------------
// Adjust the number of threads according to the current hardware
template <typename KMER_T, unsigned SIZE> void CKMC<KMER_T, SIZE>::AdjustToHardware(int32 cores)
{
	if(!cores)
	{
#ifdef THREADS_NATIVE
		cores = (int32) hardware_concurrency();
#else
		cores = (int32) thread::hardware_concurrency();
#endif
	}

	if(cores < 6)
	{
		if(Params.input_file_names.size() == 1)
			Params.n_readers = 1;
		else
			Params.n_readers = NORM(cores-1, 1, (int32) Params.input_file_names.size());

		Params.n_splitters   = NORM(cores-1, 1, 8);
		Params.n_compactors  = max(cores-1, 1);
		Params.n_sorters     = 1;
		Params.n_omp_threads = NORM(cores, 1, 8);
	}
	else if(cores <= 8)
	{
		if(Params.input_file_names.size() == 1)
			Params.n_readers = 1;
		else
			Params.n_readers = NORM(cores-1, 1, (int32) Params.input_file_names.size());
		
		Params.n_splitters   = NORM(cores-1, 1, 8);
		Params.n_compactors  = NORM(cores-1, 1, 8);
		Params.n_sorters     = 2;
		Params.n_omp_threads = NORM(cores/2+1, 1, 8);
	}
	else if(cores <= 24)
	{
		if(Params.input_file_names.size() == 1)
			Params.n_readers = 1;
		else
			Params.n_readers = MIN((int32) Params.input_file_names.size(), cores/3);

		Params.n_splitters   = (cores - Params.n_readers - 1) / 2;
		Params.n_compactors  = cores - 1 - Params.n_readers - Params.n_splitters;
		Params.n_sorters     = cores/6;
		Params.n_omp_threads = cores/Params.n_sorters;
	}	
	else
	{
		if(Params.input_file_names.size() == 1)
			Params.n_readers = 1;
		else
			Params.n_readers = MIN((int32) Params.input_file_names.size(), cores/3);

		Params.n_splitters   = (cores - Params.n_readers - 1) / 2;
		Params.n_compactors  = cores - 1 - Params.n_readers - Params.n_splitters;
		Params.n_sorters     = cores/8;
		Params.n_omp_threads = cores/Params.n_sorters;
	}	
}

//----------------------------------------------------------------------------------
// Adjust the memory limits for queues and other large data structures
template <typename KMER_T, unsigned SIZE> bool CKMC<KMER_T, SIZE>::AdjustMemoryLimits()
{
	// Memory for splitter internal buffers
	Params.max_mem_splitters = (Params.n_bins * 31 / 256 + 3) * Params.bin_part_size * sizeof(KMER_T) * Params.n_splitters;
	int64 m_rest = Params.max_mem_size - Params.max_mem_splitters;

	// Settings for memory manager of FASTQ buffers
	Params.fastq_buffer_size = 32 << 20;
	do {
		if(Params.fastq_buffer_size & (Params.fastq_buffer_size-1))
			Params.fastq_buffer_size &= Params.fastq_buffer_size - 1;
		else
			Params.fastq_buffer_size = Params.fastq_buffer_size / 2 + Params.fastq_buffer_size / 4;
		Params.mem_part_pmm_fastq = Params.fastq_buffer_size + CFastqReader::OVERHEAD_SIZE;
		Params.mem_tot_pmm_fastq  = Params.mem_part_pmm_fastq * (Params.n_readers + Params.n_splitters + 96);
	} while(Params.mem_tot_pmm_fastq > m_rest * 0.17);
	m_rest -= Params.mem_tot_pmm_fastq;

	// Subtract memory for buffers for decompression of FASTQ files
	while(Params.n_readers * Params.gzip_buffer_size > m_rest / 10)
		Params.gzip_buffer_size /= 2;
	m_rest -= Params.n_readers * Params.gzip_buffer_size;

	// Subtract memory for bin collectors internal buffers
	m_rest -= Params.n_splitters * Params.bin_part_size * sizeof(KMER_T);

	// Subtract memory for internal compactor buffers
	m_rest -= Params.n_compactors * (1ll << 23);

	// Settings for memory manager of reads
	Params.mem_part_pmm_reads = (CSplitter<KMER_T, SIZE>::MAX_LINE_SIZE + 1) * sizeof(double);
	Params.mem_tot_pmm_reads  = Params.mem_part_pmm_reads * 3 * Params.n_splitters;
	m_rest -= Params.mem_tot_pmm_reads;

	// Max. memory for single package
	Params.max_mem_storer_pkg = NORM((int64) (m_rest / (5 * Params.n_compactors + (Params.n_bins * 31 / 256 + 3) / (5*0.65))), 1ll << 20, 1ll << 26);

	// Memory for compaction of packages
	Params.max_mem_compaction = 5 * Params.n_compactors * Params.max_mem_storer_pkg;
	m_rest -= Params.max_mem_compaction;

	// Compute the lengths of compacted prefixes by BinCollectors
	uint32 count_len[3] = {0, 0, 0};
	for(uint32 j = 0; j < 3; ++j)
	{
		uint64 storage = Params.bin_part_size * sizeof(KMER_T);
		uint32 suffix_len = Params.p_k - (Params.prefix_len + (2-j));
		for(uint32 i = 0; i < min(suffix_len, (uint32) 12); ++i)
		{
			uint64 tmp_storage = 1;
			tmp_storage += Params.bin_part_size * ((suffix_len - i + 3) / 4);		// truncated kmers
			tmp_storage += ((uint64) 1 << (2*i)) * 2;								// counters

			if(tmp_storage <= storage)
			{
				storage      = tmp_storage;
				count_len[j] = i;
			}
		}
	}

	int suf_AAs       = Params.p_k - (Params.prefix_len + 2) - count_len[0];
	int pref_AAs      = count_len[0];
	int suf_As        = Params.p_k - (Params.prefix_len + 1) - count_len[1];
	int pref_As       = count_len[1];
	int suf_nAs       = Params.p_k - Params.prefix_len - count_len[2];
	int pref_nAs      = count_len[2];
	int buckets_AAs   = 1 << (2*pref_AAs);
	int buckets_As    = 1 << (2*pref_As);
	int buckets_nAs   = 1 << (2*pref_nAs);
	int size_AAs      = 2 * buckets_AAs + 2 * sizeof(uint32) + Params.bin_part_size * (suf_AAs / 4 + KMER_T::QUALITY_SIZE);
	int size_As       = 2 * buckets_As  + 2 * sizeof(uint32) + Params.bin_part_size * (suf_As  / 4 + KMER_T::QUALITY_SIZE);
	int size_nAs      = 2 * buckets_nAs + 2 * sizeof(uint32) + Params.bin_part_size * (suf_nAs / 4 + KMER_T::QUALITY_SIZE);
	Params.mem_part_pmm_bins = max(max(size_AAs, size_As), size_nAs);

	// Settings for memory managers of counters
	Params.mem_part_pmm_cnts_small = max(max(buckets_AAs, buckets_As), buckets_nAs) * sizeof(uint32);
	Params.mem_tot_pmm_cnts_small  = Params.mem_part_pmm_cnts_small * (Params.n_splitters + 2 * Params.n_compactors) * sizeof(uint32);
	m_rest -= Params.mem_tot_pmm_cnts_small;

	Params.mem_part_pmm_cnts_large = 256 * Params.mem_part_pmm_cnts_small;
	Params.mem_tot_pmm_cnts_large  = Params.mem_part_pmm_cnts_large * Params.n_compactors;
	m_rest -= Params.mem_tot_pmm_cnts_large;

	Params.mem_tot_pmm_bins = m_rest;

	// memory for storer internal buffer
	if(Params.max_mem_size >= 16ll << 30)
		Params.max_mem_storer = (int64) (Params.mem_tot_pmm_bins * 0.75);
	else
		Params.max_mem_storer = (int64) (Params.mem_tot_pmm_bins * 0.65);

	// Memory for 2nd stage
	// Settings for memory manager of radix internal buffers
	Params.mem_part_pmm_radix_buf = (256 * BUFFER_WIDTH + ALIGNMENT) * sizeof(uint64);
	Params.mem_tot_pmm_radix_buf  = Params.mem_part_pmm_radix_buf * Params.n_sorters * Params.n_omp_threads;

	Params.mem_tot_pmm_cnts_sort = Params.mem_part_pmm_cnts_large * Params.n_sorters;

	Params.max_mem_stage2	 = Params.max_mem_size - Params.mem_tot_pmm_radix_buf - Params.mem_tot_pmm_cnts_sort;

	if(Params.max_mem_storer < (1ll << 28))
		return false;

	return true;
}

//----------------------------------------------------------------------------------
// Show the settings of the KMC (in verbose mode only)
template <typename KMER_T, unsigned SIZE> void CKMC<KMER_T, SIZE>::ShowSettings()
{
	if(!Params.verbose)
		return;

	cout << "\n********** Used parameters: **********\n";

	cout << "No. of input files           : " << Params.input_file_names.size() << "\n";
	cout << "Output file name             : " << Params.output_file_name << "\n";
	cout << "No. of working directories   : " << Params.working_directories.size() << "\n";
	cout << "Input format                 : " << (Params.is_fasta ? "FASTA\n" : "FASTQ\n");

	cout << "\n";
	cout << "k-mer length                 : " << Params.kmer_len << "\n";
	cout << "Max. k-mer length            : " << MAX_K << "\n";
	cout << "Prefix length                : " << Params.prefix_len << "\n";
	cout << "Min. count threshold         : " << Params.cutoff_min << "\n";
	cout << "Max. count threshold         : " << Params.cutoff_max << "\n";
	cout << "Max. counter value           : " << Params.counter_max << "\n";
	cout << "Type of counters             : " << (Params.use_quake ? "Quake-compatibile\n" : "direct\n");
	if(Params.use_quake)
		cout << "Lowest quality value         : " << Params.lowest_quality << "\n";
	cout << "Both strands                 : " << (Params.both_strands ? "true\n" : "false\n");

	cout << "\n";
	cout << "No. of disks                 : " << Params.n_disks << "\n";
	cout << "No. of bins                  : " << Params.n_bins * 31 / 256 + 3 << "\n";
	cout << "Bin part size                : " << Params.bin_part_size << "\n";
	cout << "Input buffer size            : " << Params.fastq_buffer_size << "\n";
	cout << "\n";

	cout << "No. of readers               : " << Params.n_readers << "\n";
	cout << "No. of splitters             : " << Params.n_splitters << "\n";
	cout << "No. of compactors            : " << Params.n_compactors << "\n";
	cout << "No. of sorters               : " << Params.n_sorters << "\n";
	cout << "No. of sort. threads         : " << Params.n_omp_threads << "\n";
	cout << "\n";

	cout << "Max. mem. size               : " << setw(5) << (Params.max_mem_size >> 20) << "MB\n";
	cout << "Max. mem. per splitters      : " << setw(5) << (Params.max_mem_splitters >> 20) << "MB\n";
	cout << "Max. mem. per storer         : " << setw(5) << (Params.max_mem_storer >> 20) << "MB\n";
	cout << "Max. mem. for package comp   : " << setw(5) << (Params.max_mem_compaction >> 20) << "MB\n";
	cout << "Max. mem. for single package : " << setw(5) << (Params.max_mem_storer_pkg >> 20) << "MB\n";
	cout << "Max. mem. for 2nd stage      : " << setw(5) << (Params.max_mem_stage2 >> 20) << "MB\n";
	cout << "\n";

	cout << "Max. mem. for PMM (bin parts): " << setw(5) << (Params.mem_tot_pmm_bins >> 20) << "MB\n";
	cout << "Max. mem. for PMM (small cnt): " << setw(5) << (Params.mem_tot_pmm_cnts_small >> 20) << "MB\n";
	cout << "Max. mem. for PMM (large cnt): " << setw(5) << (Params.mem_tot_pmm_cnts_large >> 20) << "MB\n";
	cout << "Max. mem. for PMM (FASTQ)    : " << setw(5) << (Params.mem_tot_pmm_fastq >> 20) << "MB\n";
	cout << "Max. mem. for PMM (reads)    : " << setw(5) << (Params.mem_tot_pmm_reads >> 20) << "MB\n";
	cout << "Max. mem. for PMM (radix buf): " << setw(5) << (Params.mem_tot_pmm_radix_buf >> 20) << "MB\n";
	cout << "Max. mem. for PMM (sort cnts): " << setw(5) << (Params.mem_tot_pmm_cnts_sort >> 20) << "MB\n";

	cout << "\n";
}

//----------------------------------------------------------------------------------
// Run the counter
template <typename KMER_T, unsigned SIZE> bool CKMC<KMER_T, SIZE>::Process()
{
	int32 bin_id;
	FILE *file;
	string name;
	uint32 c_disk;
	uint64 size;
	uint64 n_rec;

	if(!initialized)
		return false;

	if(!AdjustMemoryLimits())
		return false;
		
	ShowSettings();

	w1.startTimer();

	// Create monitors
	Queues.mm			   = new CMemoryMonitor(Params.max_mem_stage2);
	Queues.mm_write_buffer = new CMemoryMonitor(Params.max_mem_compaction);

	Queues.dm = new CDiskMonitor(Params.n_disks);
	Queues.bo = new CBinOrdering(Params.n_bins);

	// Create queues
	Queues.input_files_queue  = new CInputFilesQueue(Params.input_file_names);
	Queues.part_queue   	  = new CPartQueue(Params.n_readers);
	Queues.bpq				  = new CBinPartQueue(Params.n_splitters);
	Queues.bd				  = new CBinDesc;
	Queues.bq				  = new CBinQueue(Params.n_disks);
	Queues.kq				  = new CKmerQueue(Params.n_bins, Params.n_sorters);
	Queues.q_package		  = new CBinPackageQueue(1);

	Queues.write_buffer_queue = new CWriteBufferQueue*[Params.n_disks];
	for(int i = 0; i < Params.n_disks; ++i)
		Queues.write_buffer_queue[i] = new CWriteBufferQueue(Params.n_compactors);

	// Create memory manager
	Queues.pmm_bins       = new CMemoryPool(Params.mem_tot_pmm_bins      , Params.mem_part_pmm_bins      );
	Queues.pmm_cnts_small = new CMemoryPool(Params.mem_tot_pmm_cnts_small, Params.mem_part_pmm_cnts_small);
	Queues.pmm_cnts_large = new CMemoryPool(Params.mem_tot_pmm_cnts_large, Params.mem_part_pmm_cnts_large);
	Queues.pmm_fastq	  = new CMemoryPool(Params.mem_tot_pmm_fastq     , Params.mem_part_pmm_fastq	 );
	Queues.pmm_reads	  = new CMemoryPool(Params.mem_tot_pmm_reads     , Params.mem_part_pmm_reads     );

	// ***** Stage 1 *****

	w_splitters.resize(Params.n_splitters);

	for(int i = 0; i < Params.n_splitters; ++i)
	{
		w_splitters[i] = new CWSplitter<KMER_T, SIZE>(Params, Queues);
#ifdef THREADS_NATIVE
		gr1_2.push_back(std::thread(std::ref(*w_splitters[i])));
#else
		gr1_2.push_back(thread(std::ref(*w_splitters[i])));
#endif
	}

	w_storer = new CWKmerBinStorer(Params, Queues);
	gr1_3.push_back(thread(std::ref(*w_storer)));

	w_compactors.resize(Params.n_compactors);
	for(int i = 0; i < Params.n_compactors; ++i)
	{
		w_compactors[i] = new CWKmerPackageCompactor(Params, Queues);
		gr1_4.push_back(thread(std::ref(*w_compactors[i])));
	}

	w_writers.resize(Params.n_disks);
	for(int i = 0; i < Params.n_disks; ++i)
	{
		w_writers[i] = new CWKmerBinWriter(Params, Queues, i);
		gr1_5.push_back(thread(std::ref(*w_writers[i])));
	}

	w_fastqs.resize(Params.n_readers);
	for(int i = 0; i < Params.n_readers; ++i)
	{
		w_fastqs[i] = new CWFastqReader(Params, Queues);
		gr1_1.push_back(thread(std::ref(*w_fastqs[i])));
	}

	for(auto p = gr1_1.begin(); p != gr1_1.end(); ++p)
		p->join();
	for(auto p = gr1_2.begin(); p != gr1_2.end(); ++p)
		p->join();

	Queues.pmm_fastq->release();
	Queues.pmm_reads->release();
	delete Queues.pmm_fastq;
	delete Queues.pmm_reads;

	for(auto p = gr1_3.begin(); p != gr1_3.end(); ++p)
		p->join();
	for(auto p = gr1_4.begin(); p != gr1_4.end(); ++p)
		p->join();
	for(auto p = gr1_5.begin(); p != gr1_5.end(); ++p)
		p->join();

	n_reads = 0;

	thread *release_thr_st1_1 = new thread([&]{
		for(int i = 0; i < Params.n_readers; ++i)
			delete w_fastqs[i];

		for(int i = 0; i < Params.n_compactors; ++i)
			delete w_compactors[i];

		for(int i = 0; i < Params.n_disks; ++i)
			delete w_writers[i];

		for(int i = 0; i < Params.n_splitters; ++i)
		{
			uint64 _n_reads;
			w_splitters[i]->GetTotal(_n_reads);
			n_reads += _n_reads;
			delete w_splitters[i];
		}

		delete w_storer;
	});

	thread *release_thr_st1_2 = new thread([&]{
		Queues.pmm_bins->release();
		delete Queues.pmm_bins;
	});

	thread *release_thr_st1_3 = new thread([&]{
		Queues.pmm_cnts_small->release();
		Queues.pmm_cnts_large->release();
		delete Queues.pmm_cnts_small;
		delete Queues.pmm_cnts_large;
	});

	release_thr_st1_1->join();
	release_thr_st1_2->join();
	release_thr_st1_3->join();
	delete release_thr_st1_1;
	delete release_thr_st1_2;
	delete release_thr_st1_3;

	w1.stopTimer();
	w2.startTimer();
	
	// ***** End of Stage 1 *****


	// Calculate LUT size
	uint32 best_lut_prefix_len = 0;
	uint64 best_mem_amount = 1ull << 62;

	for(Params.lut_prefix_len = 6; Params.lut_prefix_len < 16; ++Params.lut_prefix_len)
	{
		uint32 suffix_len = Params.kmer_len - Params.lut_prefix_len;
		if(suffix_len % 4)
			continue;

		uint64 est_suf_mem = n_reads * suffix_len;
		uint64 lut_mem     = (1ull << (2*Params.lut_prefix_len)) * sizeof(uint64);

		if(est_suf_mem + lut_mem < best_mem_amount)
		{
			best_lut_prefix_len = Params.lut_prefix_len;
			best_mem_amount     = est_suf_mem + lut_mem;
		}
	}

	Params.lut_prefix_len = best_lut_prefix_len;

	// Adjust RAM for 2nd stage
	Queues.bd->reset_reading();
	vector<int64> bin_sizes;
	while((bin_id = Queues.bd->get_next_bin()) >= 0)
	{
		Queues.bd->read(bin_id, file, name, c_disk, size, n_rec);
		bin_sizes.push_back(n_rec * 2 * sizeof(KMER_T));			// estimation of RAM for sorting bins
	}
	sort(bin_sizes.begin(), bin_sizes.end(), greater<int64>());
	int64 stage2_size = 0;
	for(int i = 0; i < 4 * Params.n_sorters; ++i)
		stage2_size += bin_sizes[i];

	stage2_size = MAX(stage2_size, 16 << 20);
	Params.max_mem_stage2 = MIN(Params.max_mem_stage2, stage2_size);

	// ***** Stage 2 *****
	Queues.bd->reset_reading();
	Queues.pmm_radix_buf  = new CMemoryPool(Params.mem_tot_pmm_radix_buf, Params.mem_part_pmm_radix_buf );
	Queues.pmm_cnts_large = new CMemoryPool(Params.mem_tot_pmm_cnts_sort, Params.mem_part_pmm_cnts_large);
	Queues.memory_bins    = new CMemoryBins(Params.max_mem_stage2, Params.n_bins);

	w_readers.resize(Params.n_disks);
	for(int i = 0; i < Params.n_disks; ++i)
	{
		w_readers[i] = new CWKmerBinReader<KMER_T, SIZE>(Params, Queues);
		gr2_1.push_back(thread(std::ref(*w_readers[i])));
	}

	w_sorters.resize(Params.n_sorters);
	for(int i = 0; i < Params.n_sorters; ++i)
	{
		w_sorters[i] = new CWKmerBinSorter<KMER_T, SIZE>(Params, Queues);
		gr2_2.push_back(thread(std::ref(*w_sorters[i])));
	}

	w_completer = new CWKmerBinCompleter(Params, Queues);
	gr2_3.push_back(thread(std::ref(*w_completer)));

	for(auto p = gr2_1.begin(); p != gr2_1.end(); ++p)
		p->join();
	for(auto p = gr2_2.begin(); p != gr2_2.end(); ++p)
		p->join();
	for(auto p = gr2_3.begin(); p != gr2_3.end(); ++p)
		p->join();

	// ***** End of Stage 2 *****
	w_completer->GetTotal(n_unique, n_cutoff_min, n_cutoff_max, n_total);

	thread *release_thr_st2_1 = new thread([&]{
		delete Queues.mm;
		delete Queues.dm;
		delete Queues.bo;

		Queues.pmm_radix_buf->release();
		Queues.pmm_cnts_large->release();
		Queues.memory_bins->release();
		delete Queues.pmm_radix_buf;
		delete Queues.pmm_cnts_large;
		delete Queues.memory_bins;
	});

	thread *release_thr_st2_2 = new thread([&]{
		for(int i = 0; i < Params.n_disks; ++i)
			delete w_readers[i];
		for(int i = 0; i < Params.n_sorters; ++i)
			delete w_sorters[i];
		delete w_completer;

		delete Queues.input_files_queue;
		delete Queues.bq;
		delete Queues.part_queue;
		delete Queues.bpq;
		delete Queues.kq;
		delete Queues.q_package;
		for(int i = 0; i < Params.n_disks; ++i)
			delete Queues.write_buffer_queue[i];
		delete[] Queues.write_buffer_queue;
	});

	// ***** Removing temporary files *****

	tmp_size = 0;
	Queues.bd->reset_reading();
	while((bin_id = Queues.bd->get_next_bin()) >= 0)
	{
		Queues.bd->read(bin_id, file, name, c_disk, size, n_rec);
		boost::filesystem::remove(boost::filesystem::path(name));
		tmp_size += size;
	}
	delete Queues.bd;

	release_thr_st2_1->join();
	release_thr_st2_2->join();
	delete release_thr_st2_1;
	delete release_thr_st2_2;

	w2.stopTimer();

	return true;
}

//----------------------------------------------------------------------------------
// Return statistics
template <typename KMER_T, unsigned SIZE> void CKMC<KMER_T, SIZE>::GetStats(double &time1, 
	double &time2, uint64 &_n_unique, uint64 &_n_cutoff_min, uint64 &_n_cutoff_max, uint64 &_n_total, uint64 &_n_reads, uint64 &_tmp_size)
{
	time1 = w1.getElapsedTime();
	time2 = w2.getElapsedTime();

	_n_unique	  = n_unique;
	_n_cutoff_min = n_cutoff_min;
	_n_cutoff_max = n_cutoff_max;
	_n_total      = n_total;
	_n_reads      = n_reads;
	_tmp_size     = tmp_size;
}

#endif

// ***** EOF
