#include "stdafx.h"
/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc
  
  Authors: Sebastian Deorowicz and Agnieszka Debudaj-Grabysz
  
  Version: 0.3
  Date   : 2012-11-29
*/

#include <fstream>
#include <string>
#include <vector>
#include <time.h>
#include <functional>
#include "timer.h"
#include "kmc.h"
#include "meta_oper.h"

using namespace std;

uint64 total_reads, total_fastq_size;

void usage();
bool parse_parameters(int argc, char *argv[]);

CKMCParams Params;

//----------------------------------------------------------------------------------
// Application class
// Template parameters:
//    * KMER_TPL - k-mer class
//    * SIZE     - maximal size of the k-mer (divided by 32)
template<template<unsigned X> class KMER_TPL, unsigned SIZE> class CApplication
{
	CApplication<KMER_TPL, SIZE-1> *app_1;
	CKMC<KMER_TPL<SIZE>, SIZE> *kmc;
	int p_k;
	bool is_selected;

public:
	CApplication(CKMCParams &Params) {
		p_k = Params.p_k;
		is_selected = p_k <= (int32) SIZE * 32 && p_k > ((int32) SIZE-1)*32;

		app_1 = new CApplication<KMER_TPL, SIZE-1>(Params);
		if(is_selected)
		{
			kmc = new CKMC<KMER_TPL<SIZE>, SIZE>;
			kmc->SetParams(Params);
		}
		else
			kmc = NULL;
	};
	~CApplication() {
		delete app_1;
		if(kmc)
			delete kmc;
	};

	void GetStats(double &time1, double &time2, uint64 &_n_unique, uint64 &_n_cutoff_min, uint64 &_n_cutoff_max, uint64 &_n_total, uint64 &_n_reads, uint64 &_tmp_size) {
		if(is_selected)
			kmc->GetStats(time1, time2, _n_unique, _n_cutoff_min, _n_cutoff_max, _n_total, _n_reads, _tmp_size);
		else
			app_1->GetStats(time1, time2, _n_unique, _n_cutoff_min, _n_cutoff_max, _n_total, _n_reads, _tmp_size);
	}

	bool Process() {
		if(is_selected)
			return kmc->Process();
		else
			return app_1->Process();
	}
};

//----------------------------------------------------------------------------------
// Specialization of the application class for the SIZE=1
template<template<unsigned X> class KMER_TPL> class CApplication<KMER_TPL, 1>
{
	CKMC<KMER_TPL<1>, 1> *kmc;
	int p_k;
	bool is_selected;

public:
	CApplication(CKMCParams &Params) {
		is_selected = Params.p_k <= 32;
		if(is_selected)
		{
			kmc = new CKMC<KMER_TPL<1>, 1>;
			kmc->SetParams(Params);
		}
		else
			kmc = NULL;
	};
	~CApplication() {
		if(kmc)
			delete kmc;
	};

	void GetStats(double &time1, double &time2, uint64 &_n_unique, uint64 &_n_cutoff_min, uint64 &_n_cutoff_max, uint64 &_n_total, uint64 &_n_reads, uint64 &_tmp_size) {
		if(is_selected)
			kmc->GetStats(time1, time2, _n_unique, _n_cutoff_min, _n_cutoff_max, _n_total, _n_reads, _tmp_size);
	}

	bool Process() {
		if(is_selected)
			return kmc->Process();
		return false;
	}
};


//----------------------------------------------------------------------------------
// Show execution options of the software
void usage()
{
	cout << "K-Mer Counter (KMC) ver. " << KMC_VER << " (" << KMC_DATE << ")\n";
	cout << "Usage:\n kmc [options] <input_file_name> <output_file_name> <working_directory> [<working_directory2> ...]\n";
	cout << " kmc [options] <@input_file_names> <output_file_name> <working_directory> [<working_directory2> ...]\n";
	cout << "Parameters:\n";
	cout << "  input_file_name - single file in FASTQ format (gziped or not)\n";
	cout << "  @input_file_names - file name with list of input files in FASTQ format (gziped or not)\n";
	cout << "Options:\n";
	cout << "  -v - verbose mode (shows all parameter settings); default: false\n";
	cout << "  -k<len> - k-mer length (k from " << MIN_K << " to " << MAX_K << "; default: 25\n";
	cout << "  -m<size> - max amount of RAM in GB (from 4 to 1024); default: 32\n";
	cout << "  -p<par> - set level of distribution (from 3 to 5); default: 4\n";
	cout << "  -f<a/q> - input in FASTA format (-fa) or FASTQ format (-fq); default: FASTQ\n";
	cout << "  -q[value] - use Quake's compatible counting with [value] representing lowest quality (default: 33)\n";
	cout << "  -ci<value> - exclude k-mers appearing less than <value> times (default: 2)\n";
	cout << "  -cs<value> - maximal value of a counter (default: 255)\n";
	cout << "  -cx<value> - exclude k-mers appearing more of than <value> times (default: 1e9)\n";
	cout << "  -b - turn off transformation of k-mers into canonical form\n";
	cout << "  -sc<value> - number of compacting threads\n";
	cout << "  -sf<value> - number of FASTQ reading threads\n";
	cout << "  -sp<value> - number of splitting threads\n";
	cout << "  -sr<value> - number of sorter threads\n";
	cout << "  -so<value> - number of threads per single sorter\n";
	cout << "Example:\n";
	cout << "kmc -k27 -m24 NA19238.fastq NA.res \\data\\kmc_tmp_dir\\\n";
	cout << "kmc -k27 -q -m24 @files.lst NA.res \\data\\kmc_tmp_dir\\\n";
}

//----------------------------------------------------------------------------------
// Parse the parameters
bool parse_parameters(int argc, char *argv[])
{
	int i;
	int tmp;

	if(argc < 4)
		return false;

	for(i = 1 ; i < argc; ++i)
	{
		if(argv[i][0] != '-')
			break;
		// Number of threads
//		if(strncmp(argv[i], "-t", 2) == 0)
//			Params.p_t = atoi(&argv[i][2]);
//		else 
		// k-mer length
		if(strncmp(argv[i], "-k", 2) == 0)
		{
			tmp = atoi(&argv[i][2]);
			if(tmp < MIN_K || tmp > MAX_K)
			{
				cout << "Wrong parameter: k must be from range <" << MIN_K << "," << MAX_K << ">\n";
				return false;
			}
			else
				Params.p_k = tmp;
		}
		// Memory limit
		else if(strncmp(argv[i], "-m", 2) == 0)
		{
			tmp = atoi(&argv[i][2]);
			if(tmp < MIN_MEM)
			{
				cout << "Wrong parameret: min memory must be at least " << MIN_MEM << "GB\n";
				return false;
			}
			else
				Params.p_m = tmp;
		}
		// Minimum counter threshold
		else if(strncmp(argv[i], "-ci", 3) == 0)
			Params.p_ci = atoi(&argv[i][3]);
		// Maximum counter threshold
		else if(strncmp(argv[i], "-cx", 3) == 0)
			Params.p_cx = atoi(&argv[i][3]);
		// Maximal counter value
		else if(strncmp(argv[i], "-cs", 3) == 0)
			Params.p_cs = atoi(&argv[i][3]);
		// Quake mode
		else if(strncmp(argv[i], "-q", 2) == 0)
		{
			Params.p_quake = true;
			if(strlen(argv[i]) > 2)
				Params.p_quality = atoi(argv[i]+2);
		}
		// Set p1
		else if(strncmp(argv[i], "-p", 2) == 0)
			Params.p_p1 = atoi(&argv[i][2]);
		// FASTA input files
		else if(strncmp(argv[i], "-fa", 3) == 0)
			Params.p_fasta = true;
		// FASTQ input files
		else if(strncmp(argv[i], "-fq", 3) == 0)
			Params.p_fasta = false;
		else if(strncmp(argv[i], "-v", 2) == 0)
			Params.p_verbose = true;
		else if(strncmp(argv[i], "-b", 2) == 0)
			Params.p_both_strands = false;
		// Number of reading threads
		else if(strncmp(argv[i], "-sf", 3) == 0)
		{
			tmp = atoi(&argv[i][3]);
			if(tmp < MIN_SF || tmp > MAX_SF)
			{
				cout << "Wrong parameter: number of reading thread must be from range <" << MIN_SF << "," << MAX_SF << ">\n";
				return false;
			}
			else
				Params.p_sf = tmp;
		}
		// Number of splitting threads
		else if(strncmp(argv[i], "-sp", 3) == 0)
		{
			tmp = atoi(&argv[i][3]);
			if(tmp < MIN_SP || tmp > MAX_SP)
			{
				cout << "Wrong parameter: number of splitting threads must be in range <" << MIN_SP << "," << MAX_SP << "<\n";
				return false;
			}
			else
				Params.p_sp = tmp;
		}
		// Number of sorting threads
		else if(strncmp(argv[i], "-so", 3) == 0)
		{
			tmp = atoi(&argv[i][3]);
			if(tmp < MIN_SO || tmp > MAX_SO)
			{	
				cout << "Wrong parameter: number of sorter threads must be in range <" << MIN_SO << "," << MAX_SO << "\n";
				return false;
			}
			else
				Params.p_so = tmp;
		}
		// Number of compactors
		else if(strncmp(argv[i], "-sc", 3) == 0)
		{
			tmp = atoi(&argv[i][3]);
			if(tmp < MIN_SC || tmp > MAX_SC)
			{
				cout << "Wrong parameter: number of compactors must be in range <" << MIN_SC << "," << MAX_SC << "\n";
				return false;
			}
			else
				Params.p_sc = tmp;
		}
		// Number of internal sorting threads (per single sorter)
		else if(strncmp(argv[i], "-sr", 3) == 0)
		{
			tmp = atoi(&argv[i][3]);
			if(tmp < MIN_SR || tmp > MAX_SR)
			{
				cout << "Wrong parameter: number of sotring threads per single sorter must be in range <" << MIN_SR << "," << MAX_SR << "\n";
				return false;
			}
			else
				Params.p_sr = tmp;
		}
	}

	if(argc - i < 3)
		return false;

	string input_file_name = string(argv[i++]);
	Params.output_file_name = string(argv[i++]);
	for(; i < argc; ++i)
		Params.working_directories.push_back(string(argv[i]));

	Params.input_file_names.clear();
	if(input_file_name[0] != '@')
		Params.input_file_names.push_back(input_file_name);
	else
	{
		ifstream in(input_file_name.c_str()+1);
		if(!in.good())
		{
			cout << "Error: No " << input_file_name.c_str()+1 << " file\n";
			return false;
		}

		string s;
		while(getline(in, s))
			if(s != "")
				Params.input_file_names.push_back(s);

		in.close();
		random_shuffle(Params.input_file_names.begin(), Params.input_file_names.end());
	}

	return true;
}

//----------------------------------------------------------------------------------
// Main function
int _tmain(int argc, _TCHAR* argv[])
{
	CStopWatch w0, w1;
	double time1, time2;
	uint64 n_unique, n_cutoff_min, n_cutoff_max, n_total, n_reads, tmp_size;

	omp_set_num_threads(1);

#ifdef WIN32
	_setmaxstdio(2040);
#endif

	if(!parse_parameters(argc, argv))
	{
		usage();
		return 0;
	}

	if(Params.p_quake)
	{
		CApplication<CKmerQuake, KMER_WORDS> *app = new CApplication<CKmerQuake, KMER_WORDS>(Params);

		if(!app->Process())
		{
			cout << "Not enough memory or some other error\n";
			delete app;
			return 0;
		}
		app->GetStats(time1, time2, n_unique, n_cutoff_min, n_cutoff_max, n_total, n_reads, tmp_size);
		delete app;
	}
	else
	{
		CApplication<CKmer, KMER_WORDS> *app = new CApplication<CKmer, KMER_WORDS>(Params);

		if(!app->Process())
		{
			cout << "Not enough memory or some other error\n";
			delete app;
			return 0;
		}
		app->GetStats(time1, time2, n_unique, n_cutoff_min, n_cutoff_max, n_total, n_reads, tmp_size);
		delete app;
	}

	cout << "1st stage: " << time1 << "s\n";
	cout << "2nd stage: " << time2  << "s\n";
	cout << "Total    : " << (time1+time2) << "s\n";
	cout << "Tmp size : " << tmp_size / (1 << 20) << "MB\n";
	cout << "\nStats:\n";
	cout << "   No. of k-mers below min. threshold : " << setw(12) << n_cutoff_min << "\n";
	cout << "   No. of k-mers above max. threshold : " << setw(12) << n_cutoff_max << "\n";
	cout << "   No. of unique k-mers               : " << setw(12) << n_unique << "\n";
	cout << "   No. of unique counted k-mers       : " << setw(12) << n_unique-n_cutoff_min-n_cutoff_max << "\n";
	cout << "   Total no. of k-mers                : " << setw(12) << n_total << "\n";
	cout << "   Total no. of reads                 : " << setw(12) << n_reads << "\n";

	return 0;
}

// ***** EOF
