/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc

  Authors: Sebastian Deorowicz and Agnieszka Debudaj-Grabysz

  Version: 0.3
  Date   : 2012-11-29
*/

#ifndef _PARAMS_H
#define _PARAMS_H

#include "defs.h"
#include "queues.h"
#include <vector>
#include <string>

using namespace std;

// Structure for passing KMC parameters
struct CKMCParams {
	// Input parameters
	int p_m;							// max. total RAM usage
	int p_k;							// k-mer length
	int p_t;							// no. of threads
	int p_sf;							// no. of reading threads
	int p_sp;							// no. of splitting threads
	int p_so;							// no. of OpenMP threads for sorting
	int p_sr;							// no. of sorting threads
	int p_sc;							// no. of compacting threads
	int p_ci;							// do not count k-mers occurring less than
	int p_cx;							// do not count k-mers occurring more than
	int p_cs;							// maximal counter value
	bool p_quake;						// use Quake-compatibile counting
	int p_quality;						// lowest quality
	bool p_fasta;						// input in FASTA format
	bool p_verbose;						// verbose mode
	bool p_both_strands;				// compute canonical k-mer representation
	int p_p1;							// p1 length

	// File names
	vector<string> input_file_names;
	string output_file_name;
	vector<string> working_directories;
	bool is_fasta;

	uint32 lut_prefix_len;
	
	uint32 KMER_T_size;

	// Memory sizes
	int64 max_mem_size;				// maximum amount of memory to be used in GBs;	default: 30GB
	int64 max_mem_splitters;		// maximum amount of memory for splitters buffers
	int64 max_mem_storer;			// maximum amount of memory for internal buffers of KmerStorer
	int64 max_mem_stage2;			// maximum amount of memory in stage 2
	int64 max_mem_storer_pkg;		// maximum amount of memory for single package
	int64 max_mem_compaction;		// maximum amount of memory for compaction of packages

	int64 mem_tot_pmm_bins;			// maximal amount of memory per pool memory manager (PMM) of bin parts
	int64 mem_part_pmm_bins;		// maximal amount of memory per single part of memory maintained by PMM of bin parts
	int64 mem_tot_pmm_cnts_small;	
	int64 mem_part_pmm_cnts_small;
	int64 mem_tot_pmm_cnts_large;	
	int64 mem_part_pmm_cnts_large;
	int64 mem_tot_pmm_fastq;
	int64 mem_part_pmm_fastq;
	int64 mem_part_pmm_reads;
	int64 mem_tot_pmm_reads;
	int64 mem_part_pmm_radix_buf;
	int64 mem_tot_pmm_radix_buf;
	int64 mem_part_pmm_cnts_sort;
	int64 mem_tot_pmm_cnts_sort;

	bool verbose;

	int kmer_len;			// kmer length
	int prefix_len;			// prefix length; 3-4, 4-5, or 5-6
	int cutoff_min;			// exclude k-mers appearing less than times
	int cutoff_max;			// exclude k-mers appearing more than times
	int counter_max;		// maximal counter value
	bool use_quake;			// use Quake's counting based on qualities
	int lowest_quality;		// lowest quality value	    
	bool both_strands;		// find canonical representation of each k-mer

	int n_disks;
	int n_bins;				// number of bins; fixed: 448
	int bin_part_size;		// size of a bin part; fixed: 2^15
	int fastq_buffer_size;	// size of FASTQ file buffer; fixed: 2^23

	int n_threads;			// number of cores
	int n_readers;			// number of FASTQ readers; default: 1
	int n_splitters;		// number of splitters; default: 1
	int n_sorters;			// number of sorters; default: 1
	int n_omp_threads;		// number of OMP threads per each sorter
	int n_compactors;		// number of bin part compactors

	uint32 gzip_buffer_size;
	uint32 bzip2_buffer_size;

	CKMCParams()
	{
		p_m = 8;
		p_k = 25;
		p_t = 0;
		p_sf = 0;
		p_sp = 0;
		p_so = 0;
		p_sr = 0;
		p_sc = 0;
		p_ci = 2;
		p_cx = 1000000000;
		p_cs = 255;
		p_quake = false;
		p_quality = 33;
		p_fasta = false;
		p_verbose = false;
		p_both_strands = true;
		p_p1 = 4;

		gzip_buffer_size  = 64 << 20;
		bzip2_buffer_size = 64 << 20;
	}
};

// Structure for passing KMC queues and monitors to threads
struct CKMCQueues 
{
	// Memory and disk monitors
	CMemoryMonitor *mm;
	CMemoryMonitor *mm_write_buffer;

	CDiskMonitor *dm;
	CBinOrdering *bo;

	// Queues
	CInputFilesQueue *input_files_queue;
	CPartQueue *part_queue;
	CWriteBufferQueue **write_buffer_queue;
	CBinPartQueue *bpq;
	CBinDesc *bd;
	CBinQueue *bq;
	CKmerQueue *kq;
	CBinPackageQueue *q_package;

	CMemoryPool *pmm_bins, *pmm_cnts_small, *pmm_cnts_large, *pmm_fastq, *pmm_reads, *pmm_radix_buf;
	CMemoryBins *memory_bins;

	CKMCQueues() {}
};

#endif

// ***** EOF
