/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc

  Authors: Sebastian Deorowicz and Agnieszka Debudaj-Grabysz

  Version: 0.3
  Date   : 2012-11-29
*/

#ifndef _SPLITTER_H
#define _SPLITTER_H

#include "defs.h"
#include "kmer.h"
#include "kb_storer.h"
#include "kb_collector.h"
#include "kb_reader.h"
#include "kb_sorter.h"
#include "kb_completer.h"
#include "queues.h"
#include <stdio.h>
#include <iostream>
#include <vector>
using namespace std;


//************************************************************************************************************
template <typename KMER_T, unsigned SIZE> class CSplitter_Impl;

//************************************************************************************************************
// CSplitter class - splits kmers into bins according to their prefix
//************************************************************************************************************
template <typename KMER_T, unsigned SIZE> class CSplitter {
	CMemoryMonitor *mm;

	uchar *part;
	uint64 part_size, part_pos;
	CKmerBinCollector<KMER_T, SIZE> **bins;
	CBinPartQueue *bin_part_queue;
	CBinDesc *bd;
	CMemoryPool *pmm_reads;
	int64 mem_part_pmm_bins;

	char codes[256];
	bool use_quake;
	bool is_fasta;
	int lowest_quality;
	bool both_strands;

	uint32 kmer_len;
	uint32 prefix_len;
	uint32 n_bins;

	uint64 n_reads;

	inline bool GetSeq(char *seq, uint32 &seq_size);
	inline bool GetSeq(char *seq, char *quals, uint32 &seq_size);
	inline void PushBinToQueue(int bin_no, bool last_one);

	friend class CSplitter_Impl<KMER_T, SIZE>;

public:
	static uint32 MAX_LINE_SIZE;

	CSplitter(CKMCParams &Params, CKMCQueues &Queues);
	~CSplitter();

	bool ProcessReads(uchar *_part, uint64 _part_size);
	void Complete();

	void GetTotal(uint64 &_n_reads);
};

template <typename KMER_T, unsigned SIZE> uint32 CSplitter<KMER_T, SIZE>::MAX_LINE_SIZE = 1 << 14;


//************************************************************************************************************
// Implementation of ProcessReads and Complete methods for various types and sizes of kmer class
//************************************************************************************************************
template <typename KMER_T, unsigned SIZE> class CSplitter_Impl {
public: 
	static bool ProcessReads(CSplitter<KMER_T, SIZE> &ptr, uchar *_part, uint64 _part_size);
	static void Complete(CSplitter<KMER_T, SIZE> &ptr);
};

template <unsigned SIZE> class CSplitter_Impl<CKmer<SIZE>, SIZE> {
public: 
	static bool ProcessReads(CSplitter<CKmer<SIZE>, SIZE> &ptr, uchar *_part, uint64 _part_size);
	static void Complete(CSplitter<CKmer<SIZE>, SIZE> &ptr);
};

template <unsigned SIZE> class CSplitter_Impl<CKmerQuake<SIZE>, SIZE> {
  static double prob_qual[94];
  static double inv_prob_qual[94];
  static double MIN_PROB_QUAL_VALUE;

public: 
	static bool ProcessReads(CSplitter<CKmerQuake<SIZE>, SIZE> &ptr, uchar *_part, uint64 _part_size);
	static void Complete(CSplitter<CKmerQuake<SIZE>, SIZE> &ptr);
};

// K-mers with probability less than MIN_PROB_QUAL_VALUE will not be counted
template <unsigned SIZE> double CSplitter_Impl<CKmerQuake<SIZE>, SIZE>::MIN_PROB_QUAL_VALUE = 0.0000;

template <unsigned SIZE> double CSplitter_Impl<CKmerQuake<SIZE>, SIZE>::prob_qual[94] = {
    0.2500000000000000, 0.2500000000000000, 0.3690426555198070, 0.4988127663727280, 
	0.6018928294465030, 0.6837722339831620, 0.7488113568490420, 0.8004737685031120, 
	0.8415106807538890, 0.8741074588205830, 0.9000000000000000, 0.9205671765275720, 
	0.9369042655519810, 0.9498812766372730, 0.9601892829446500, 0.9683772233983160, 
	0.9748811356849040, 0.9800473768503110, 0.9841510680753890, 0.9874107458820580, 
	0.9900000000000000, 0.9920567176527570, 0.9936904265551980, 0.9949881276637270, 
	0.9960189282944650, 0.9968377223398320, 0.9974881135684900, 0.9980047376850310, 
	0.9984151068075390, 0.9987410745882060, 0.9990000000000000, 0.9992056717652760, 
	0.9993690426555200, 0.9994988127663730, 0.9996018928294460, 0.9996837722339830, 
	0.9997488113568490, 0.9998004737685030, 0.9998415106807540, 0.9998741074588210, 
	0.9999000000000000, 0.9999205671765280, 0.9999369042655520, 0.9999498812766370, 
	0.9999601892829450, 0.9999683772233980, 0.9999748811356850, 0.9999800473768500, 
	0.9999841510680750, 0.9999874107458820, 0.9999900000000000, 0.9999920567176530, 
	0.9999936904265550, 0.9999949881276640, 0.9999960189282940, 0.9999968377223400, 
	0.9999974881135680, 0.9999980047376850, 0.9999984151068080, 0.9999987410745880, 
	0.9999990000000000, 0.9999992056717650, 0.9999993690426560, 0.9999994988127660, 
	0.9999996018928290, 0.9999996837722340, 0.9999997488113570, 0.9999998004737680, 
	0.9999998415106810, 0.9999998741074590, 0.9999999000000000, 0.9999999205671770, 
	0.9999999369042660, 0.9999999498812770, 0.9999999601892830, 0.9999999683772230, 
	0.9999999748811360, 0.9999999800473770, 0.9999999841510680, 0.9999999874107460, 
	0.9999999900000000, 0.9999999920567180, 0.9999999936904270, 0.9999999949881280, 
	0.9999999960189280, 0.9999999968377220, 0.9999999974881140, 0.9999999980047380, 
	0.9999999984151070, 0.9999999987410750, 0.9999999990000000, 0.9999999992056720, 
	0.9999999993690430, 0.9999999994988130};

template <unsigned SIZE> double CSplitter_Impl<CKmerQuake<SIZE>, SIZE>::inv_prob_qual[94] = {
	4.0000000000000000, 4.0000000000000000, 2.7097138638119600, 2.0047602375372500,
	1.6614253419825500, 1.4624752955742600, 1.3354498310601800, 1.2492601748462100,
	1.1883390465158700, 1.1440241012807300, 1.1111111111111100, 1.0862868300084900,
	1.0673449110735400, 1.0527631448218000, 1.0414613220148200, 1.0326554320337200,
	1.0257660789563300, 1.0203588353185700, 1.0161041657513100, 1.0127497641386300,
	1.0101010101010100, 1.0080068832818700, 1.0063496369454600, 1.0050371177272600,
	1.0039969839853900, 1.0031723093832600, 1.0025182118938000, 1.0019992513458400, 
	1.0015874090662800, 1.0012605123027600, 1.0010010010010000, 1.0007949596936500,
	1.0006313557030000, 1.0005014385482300, 1.0003982657229900, 1.0003163277976500,
	1.0002512517547400, 1.0001995660501600, 1.0001585144420900, 1.0001259083921100,
	1.0001000100010000, 1.0000794391335500, 1.0000630997157700, 1.0000501212353700,
	1.0000398123020100, 1.0000316237766300, 1.0000251194952900, 1.0000199530212600,
	1.0000158491831200, 1.0000125894126100, 1.0000100001000000, 1.0000079433454400,
	1.0000063096132600, 1.0000050118974600, 1.0000039810875500, 1.0000031622876600,
	1.0000025118927400, 1.0000019952663000, 1.0000015848957000, 1.0000012589270000,
	1.0000010000010000, 1.0000007943288700, 1.0000006309577400, 1.0000005011874800,
	1.0000003981073300, 1.0000003162278700, 1.0000002511887100, 1.0000001995262700, 
	1.0000001584893400, 1.0000001258925600, 1.0000001000000100, 1.0000000794328300,
	1.0000000630957400, 1.0000000501187300, 1.0000000398107200, 1.0000000316227800,
	1.0000000251188600, 1.0000000199526200, 1.0000000158489300, 1.0000000125892500,
	1.0000000100000000, 1.0000000079432800, 1.0000000063095700, 1.0000000050118700,
	1.0000000039810700, 1.0000000031622800, 1.0000000025118900, 1.0000000019952600,
	1.0000000015848900, 1.0000000012589300, 1.0000000010000000, 1.0000000007943300,
	1.0000000006309600, 1.0000000005011900};


//----------------------------------------------------------------------------------
// Return a single record from FASTA/FASTQ data
template <typename KMER_T, unsigned SIZE> bool CSplitter<KMER_T, SIZE>::GetSeq(char *seq, uint32 &seq_size)
{
	uchar c;
	uint32 pos = 0;

	if(is_fasta)
	{
		// Title
		if(part_pos >= part_size)
			return false;
		c = part[part_pos++];
		if(c != '>')
			return false;
		for(; part_pos < part_size;)
		{
			c = part[part_pos++];
			if(c < 32)					// newliners
				break;
		}
		if(part_pos >= part_size)
			return false;

		c = part[part_pos++];
		if(c >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return false;

		// Sequence
		for(; part_pos < part_size;)
		{
			c = part[part_pos++];
			if(c < 32)					// newliners
				break;
			seq[pos++] = codes[c];
		}
		seq_size = pos;

		if(part_pos >= part_size)
			return true;

		if(part[part_pos++] >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return true;
	}
	else
	{
		// Title
		if(part_pos >= part_size)
			return false;
		c = part[part_pos++];
		if(c != '@')
			return false;
		for(; part_pos < part_size;)
		{
			c = part[part_pos++];
			if(c < 32)					// newliners
				break;
		}
		if(part_pos >= part_size)
			return false;

		c = part[part_pos++];
		if(c >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return false;

		// Sequence
		for(; part_pos < part_size;)
		{
			c = part[part_pos++];
			if(c < 32)					// newliners
				break;
			seq[pos++] = codes[c];
		}
		if(part_pos >= part_size)
			return false;

		c = part[part_pos++];
		if(c >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return false;

		// Plus
		c = part[part_pos++];
		if(part_pos >= part_size)
			return false;
		if(c != '+')
			return false;
		for(; part_pos < part_size;)
		{
			c = part[part_pos++];
			if(c < 32)					// newliners
				break;
		}
		if(part_pos >= part_size)
			return false;

		c = part[part_pos++];
		if(c >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return false;

		// Quality
		part_pos += pos;
		if(part_pos >= part_size)
			return false;
		c = part[part_pos++];
		seq_size = pos;

		if(part_pos >= part_size)
			return true;

		if(part[part_pos++] >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return true;
	}

	return (c == '\n' || c == '\r');
}

//----------------------------------------------------------------------------------
// Return a single record with quality codes from FASTA/FASTQ data
template <typename KMER_T, unsigned SIZE> bool CSplitter<KMER_T, SIZE>::GetSeq(char *seq, char *quals, uint32 &seq_size)
{
	uchar c;
	uint32 pos = 0;

	if(is_fasta)
	{
		return false;		// FASTA file does not store quality values
	}
	else
	{
		// Title
		if(part_pos >= part_size)
			return false;
		c = part[part_pos++];
		if(c != '@')
			return false;
		for(; part_pos < part_size;)
		{
			c = part[part_pos++];
			if(c < 32)					// newliners
				break;
		}
		if(part_pos >= part_size)
			return false;

		c = part[part_pos++];
		if(c >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return false;

		// Sequence
		for(; part_pos < part_size;)
		{
			c = part[part_pos++];
			if(c < 32)					// newliners
				break;
			seq[pos++] = codes[c];
		}
		if(part_pos >= part_size)
			return false;

		c = part[part_pos++];
		if(c >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return false;

		// Plus
		c = part[part_pos++];
		if(part_pos >= part_size)
			return false;
		if(c != '+')
			return false;
		for(; part_pos < part_size;)
		{
			c = part[part_pos++];
			if(c < 32)					// newliners
				break;
		}
		if(part_pos >= part_size)
			return false;

		c = part[part_pos++];
		if(c >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return false;

		// Quality
		copy(part+part_pos, part+part_pos+pos, quals);

		part_pos += pos;
		if(part_pos >= part_size)
			return false;
		c = part[part_pos++];
		seq_size = pos;

		if(part_pos >= part_size)
			return true;

		if(part[part_pos++] >= 32)
			part_pos--;
		else if(part_pos >= part_size)
			return true;
	}

	return (c == '\n' || c == '\r');
}

//----------------------------------------------------------------------------------
// Push a full bin into a queue of bins
template <typename KMER_T, unsigned SIZE> void CSplitter<KMER_T, SIZE>::PushBinToQueue(int bin_no, bool last_one)
{
	uchar *bin_part;
	uint64 bin_part_size;
	uint64 n_rec;

	if(bins[bin_no]->GetBinPart(bin_part, bin_part_size, n_rec, last_one))
	{
		bin_part_queue->push(bin_no, bin_part, (uint32) bin_part_size, (uint32) mem_part_pmm_bins);
		bd->insert(bin_no, NULL, "", 0, bin_part_size, n_rec);
	}
}

//----------------------------------------------------------------------------------
// Assigns queues and monitors
template <typename KMER_T, unsigned SIZE> CSplitter<KMER_T, SIZE>::CSplitter(CKMCParams &Params, CKMCQueues &Queues)
{
	mm             = Queues.mm;
	is_fasta       = Params.is_fasta;
	use_quake	   = Params.use_quake;
	lowest_quality = Params.lowest_quality;
	both_strands   = Params.both_strands;

	bin_part_queue = Queues.bpq;
	bd			   = Queues.bd;
	pmm_reads	   = Queues.pmm_reads;

	kmer_len		  = Params.kmer_len;
	prefix_len		  = Params.prefix_len;
	n_bins            = Params.n_bins;
	mem_part_pmm_bins = Params.mem_part_pmm_bins;

	part = NULL;

	// Prepare encoding of symbols
	for(int i = 0; i < 256; ++i)
		codes[i] = -1;
	codes['A'] = codes['a'] = 0;
	codes['C'] = codes['c'] = 1;
	codes['G'] = codes['g'] = 2;
	codes['T'] = codes['t'] = 3;

	uint32 buffer_size = Params.bin_part_size;
	uint32 count_len;

	// Create objects for all bin
	bins = new CKmerBinCollector<KMER_T, SIZE>*[n_bins];

	for(uint32 i = 0; i < n_bins; ++i)
		if(i < 4)
		{
			bins[i] = new CKmerBinCollector<KMER_T, SIZE>(Params, Queues, i, buffer_size, prefix_len+2, kmer_len);
			bins[i]->GetCompressionParams(count_len);
			bd->insert(i, NULL, "", 0, 0, 0, buffer_size, prefix_len+2, kmer_len, count_len);
		}
		else if(i < n_bins/4 + n_bins/16 && i % 4 == 0)
		{
			bins[i] = new CKmerBinCollector<KMER_T, SIZE>(Params, Queues, i, buffer_size, prefix_len+1, kmer_len);
			bins[i]->GetCompressionParams(count_len);
			bd->insert(i, NULL, "", 0, 0, 0, buffer_size, prefix_len+1, kmer_len, count_len);
		}
		else if(i % 16 == 0)
		{
			bins[i] = new CKmerBinCollector<KMER_T, SIZE>(Params, Queues, i, buffer_size, prefix_len, kmer_len);
			bins[i]->GetCompressionParams(count_len);
			bd->insert(i, NULL, "", 0, 0, 0, buffer_size, prefix_len, kmer_len, count_len);
		}
		else
			bins[i] = NULL;

	n_reads = 0;
}

//----------------------------------------------------------------------------------
// Release memory
template <typename KMER_T, unsigned SIZE> CSplitter<KMER_T, SIZE>::~CSplitter()
{
	if(bins)
		for(uint32 i = 0; i < n_bins; ++i)
			if(bins[i])
				delete bins[i];
	delete[] bins;
}

//----------------------------------------------------------------------------------
// Finish the processing of input file
template <typename KMER_T, unsigned SIZE> void CSplitter<KMER_T, SIZE>::Complete()
{
	return CSplitter_Impl<KMER_T, SIZE>::Complete(*this);
}

//----------------------------------------------------------------------------------
// Process the reads from the given FASTQ file part
template <typename KMER_T, unsigned SIZE> bool CSplitter<KMER_T, SIZE>::ProcessReads(uchar *_part, uint64 _part_size)
{
	return CSplitter_Impl<KMER_T, SIZE>::ProcessReads(*this, _part, _part_size);
}

//----------------------------------------------------------------------------------
// Return the number of reads processed by splitter
template <typename KMER_T, unsigned SIZE> void CSplitter<KMER_T, SIZE>::GetTotal(uint64 &_n_reads)
{
	_n_reads = n_reads;
}


//************************************************************************************************************
// Implementation of specific splitter methods for various types and sizes of kmers
//************************************************************************************************************

//----------------------------------------------------------------------------------
// Process the reads from the given FASTQ file part
template <unsigned SIZE> bool CSplitter_Impl<CKmer<SIZE>, SIZE>::ProcessReads(CSplitter<CKmer<SIZE>, SIZE> &ptr, uchar *_part, uint64 _part_size)
{
	ptr.part      = _part;
	ptr.part_size = _part_size;
	ptr.part_pos  = 0;

	char *seq;
	uint32 seq_size;
	int omit_next_n_kmers;
	CKmer<SIZE> kmer_str, kmer_rev, kmer_can;
	uint32 i;
	CKmer<SIZE> kmer_mask;
	uint32 bin_shift = (2*ptr.kmer_len - 2*ptr.prefix_len);
	
	ptr.pmm_reads->reserve(seq);

	kmer_mask.set_n_1(2*ptr.kmer_len);

	uint32 kmer_len_shift  = (ptr.kmer_len-1)*2;

	if(ptr.both_strands)
		while(ptr.GetSeq(seq, seq_size))
		{
			ptr.n_reads++;

			// Init k-mer
			kmer_str.clear();
			kmer_rev.clear();

			// Process first k-1 symbols of a read
			uint32 str_pos = kmer_len_shift-2;
			uint32 rev_pos = 2;

			omit_next_n_kmers = 0;

			for(i = 0; i < ptr.kmer_len-1; ++i, str_pos-=2, rev_pos+=2)
			{
				if(seq[i] < 0)
				{
					seq[i] = 0;
					omit_next_n_kmers = i+1;
				}
				kmer_str.set_2bits(seq[i], str_pos);
				kmer_rev.set_2bits(3-seq[i], rev_pos);
			}

			// Process next part of a read
			for(; i < seq_size; ++i)
			{
				if(seq[i] < 0)		// N in a read
				{
					seq[i] = 0;
					omit_next_n_kmers = ptr.kmer_len;		// Mark how many symbols to ommit to get the next kmer without any N
				}
				kmer_str.SHL_insert_2bits(seq[i]);
				kmer_str.mask(kmer_mask);
				kmer_rev.SHR_insert_2bits(3 - seq[i], kmer_len_shift);
			
				// If necessary ommit next symbols
				if(omit_next_n_kmers > 0)
				{
					omit_next_n_kmers--;
					continue;
				}

				// Find canonical kmer representation
				kmer_can = (kmer_str < kmer_rev) ? kmer_str : kmer_rev;

				uint32 bin_no = (uint32) kmer_can.remove_suffix(bin_shift-4);

				if(bin_no < 4)
					;
				else if(bin_no < ptr.n_bins/4 + ptr.n_bins/16)
					bin_no &= ~3u;
				else
					bin_no &= ~0xfu;

				// Put kmer to the proper bin
				if(ptr.bins[bin_no]->PutKmer(kmer_can))
					ptr.PushBinToQueue(bin_no, false);
			}
		}
	else
		while(ptr.GetSeq(seq, seq_size))
		{
			ptr.n_reads++;

			// Init k-mer
			kmer_str.clear();

			// Process first k-1 symbols of a read
			uint32 str_pos = kmer_len_shift-2;

			omit_next_n_kmers = 0;

			for(i = 0; i < ptr.kmer_len-1; ++i, str_pos-=2)
			{
				if(seq[i] < 0)
				{
					seq[i] = 0;
					omit_next_n_kmers = i+1;
				}
				kmer_str.set_2bits(seq[i], str_pos);
			}

			// Process next part of a read
			for(; i < seq_size; ++i)
			{
				if(seq[i] < 0)		// N in a read
				{
					seq[i] = 0;
					omit_next_n_kmers = ptr.kmer_len;		// Mark how many symbols to ommit to get the next kmer without any N
				}
				kmer_str.SHL_insert_2bits(seq[i]);
				kmer_str.mask(kmer_mask);
			
				// If necessary ommit next symbols
				if(omit_next_n_kmers > 0)
				{
					omit_next_n_kmers--;
					continue;
				}

				uint32 bin_no = (uint32) kmer_str.remove_suffix(bin_shift-4);

				if(bin_no < 4)
					;
				else if(bin_no < ptr.n_bins/4 + ptr.n_bins/16)
					bin_no &= ~3u;
				else
					bin_no &= ~0xfu;
				// Put kmer to the proper bin
				if(ptr.bins[bin_no]->PutKmer(kmer_str))
					ptr.PushBinToQueue(bin_no, false);
			}
		}
	putchar('*');
	fflush(stdout);

	ptr.pmm_reads->free(seq);

	return true;
}

//----------------------------------------------------------------------------------
// Process the reads from the given FASTQ file part
template <unsigned SIZE> void CSplitter_Impl<CKmer<SIZE>, SIZE>::Complete(CSplitter<CKmer<SIZE>, SIZE> &ptr)
{
	// Push data from bin to a queue of bin parts
	if(ptr.bins)
		for(uint32 i = 0; i < ptr.n_bins; ++i)
			if(ptr.bins[i])
				ptr.PushBinToQueue(i, true);
}

//----------------------------------------------------------------------------------
// Process the reads from the given FASTQ file part
template <unsigned SIZE> bool CSplitter_Impl<CKmerQuake<SIZE>, SIZE>::ProcessReads(CSplitter<CKmerQuake<SIZE>, SIZE> &ptr, uchar *_part, uint64 _part_size)
{
	ptr.part      = _part;
	ptr.part_size = _part_size;
	ptr.part_pos  = 0;

	char *seq; 
	char *quals;
	double *raw_inv_probs; 

	ptr.pmm_reads->reserve(seq);
	ptr.pmm_reads->reserve(quals);
	ptr.pmm_reads->reserve(raw_inv_probs);

	double *inv_probs = raw_inv_probs+1;
	inv_probs[-1] = 1.0;			// !!! Correct

	uint32 seq_size;
	int omit_next_n_kmers;
	CKmerQuake<SIZE> kmer_str, kmer_rev, kmer_can;
	double kmer_prob;

	uint32 i;
	CKmerQuake<SIZE> kmer_mask;
	uint32 bin_shift = (2*ptr.kmer_len - 2*ptr.prefix_len);
	
	kmer_mask.set_n_1(2*ptr.kmer_len);

	uint32 kmer_len_shift  = (ptr.kmer_len-1)*2;

	if(ptr.both_strands)
		while(ptr.GetSeq(seq, quals, seq_size))
		{
			ptr.n_reads++;

			// Init k-mer
			kmer_str.clear();
			kmer_rev.clear();

			// Process first k-1 symbols of a read
			uint32 str_pos = kmer_len_shift-2;
			uint32 rev_pos = 2;

			omit_next_n_kmers = 0;
			kmer_prob = 1.0;

			for(i = 0; i < ptr.kmer_len-1; ++i, str_pos-=2, rev_pos+=2)
			{
				if(seq[i] < 0)
				{
					seq[i] = 0;
					omit_next_n_kmers = i+1;

				}
					inv_probs[i] = inv_prob_qual[quals[i] - ptr.lowest_quality];

				kmer_str.set_2bits(seq[i], str_pos);
				kmer_rev.set_2bits(3-seq[i], rev_pos);
				kmer_prob *= prob_qual[quals[i] - ptr.lowest_quality];
			}

			// Process next part of a read
			for(; i < seq_size; ++i)
			{
				if(seq[i] < 0)		// N in a read
				{
					seq[i] = 0;
					omit_next_n_kmers = ptr.kmer_len;		// Mark how many symbols to ommit to get the next kmer without any N
				}
				inv_probs[i] = inv_prob_qual[quals[i] - ptr.lowest_quality];

				kmer_str.SHL_insert_2bits(seq[i]);
				kmer_str.mask(kmer_mask);
				kmer_rev.SHR_insert_2bits(3 - seq[i], kmer_len_shift);
				kmer_prob *= prob_qual[quals[i] - ptr.lowest_quality] * inv_probs[(int)i - (int)ptr.kmer_len];

				// If necessary ommit next symbols
				if(omit_next_n_kmers > 0)
				{
					omit_next_n_kmers--;
					continue;
				}

				if(kmer_prob < MIN_PROB_QUAL_VALUE)
					continue;

				// Find canonical kmer representation
				kmer_can = (kmer_str < kmer_rev) ? kmer_str : kmer_rev;

				uint32 bin_no = (uint32) kmer_can.remove_suffix(bin_shift-4);

				if(bin_no < 4)
					;
				else if(bin_no < ptr.n_bins/4 + ptr.n_bins/16)
					bin_no &= ~3u;
				else
					bin_no &= ~0xfu;

				kmer_can.quality = (float) kmer_prob;

				if(ptr.bins[bin_no]->PutKmer(kmer_can))
					ptr.PushBinToQueue(bin_no, false);
			}
		}
	else
		while(ptr.GetSeq(seq, quals, seq_size))
		{
			ptr.n_reads++;

			// Init k-mer
			kmer_str.clear();

			// Process first k-1 symbols of a read
			uint32 str_pos = kmer_len_shift-2;

			omit_next_n_kmers = 0;
			kmer_prob = 1.0;

			for(i = 0; i < ptr.kmer_len-1; ++i, str_pos-=2)
			{
				if(seq[i] < 0)
				{
					seq[i] = 0;
					omit_next_n_kmers = i+1;

				}
				inv_probs[i] = inv_prob_qual[quals[i] - ptr.lowest_quality];

				kmer_str.set_2bits(seq[i], str_pos);
				kmer_prob *= prob_qual[quals[i] - ptr.lowest_quality];
			}

			// Process next part of a read
			for(; i < seq_size; ++i)
			{
				if(seq[i] < 0)		// N in a read
				{
					seq[i] = 0;
					omit_next_n_kmers = ptr.kmer_len;		// Mark how many symbols to ommit to get the next kmer without any N
				}
					inv_probs[i] = inv_prob_qual[quals[i] - ptr.lowest_quality];

				kmer_str.SHL_insert_2bits(seq[i]);
				kmer_str.mask(kmer_mask);
				kmer_prob *= prob_qual[quals[i] - ptr.lowest_quality] * inv_probs[(int)i - (int)ptr.kmer_len];

				// If necessary ommit next symbols
				if(omit_next_n_kmers > 0)
				{
					omit_next_n_kmers--;
					continue;
				}

				if(kmer_prob < MIN_PROB_QUAL_VALUE)
					continue;

				uint32 bin_no = (uint32) kmer_str.remove_suffix(bin_shift-4);

				if(bin_no < 4)
					;
				else if(bin_no < ptr.n_bins/4 + ptr.n_bins/16)
					bin_no &= ~3u;
				else
					bin_no &= ~0xfu;

				kmer_can.quality = (float) kmer_prob;

				if(ptr.bins[bin_no]->PutKmer(kmer_str))
					ptr.PushBinToQueue(bin_no, false);
			}
		}

	putchar('*');
	fflush(stdout);

	ptr.pmm_reads->free(seq);
	ptr.pmm_reads->free(quals);
	ptr.pmm_reads->free(raw_inv_probs);

	return true;
}

//----------------------------------------------------------------------------------
// Process the reads from the given FASTQ file part
template <unsigned SIZE> void CSplitter_Impl<CKmerQuake<SIZE>, SIZE>::Complete(CSplitter<CKmerQuake<SIZE>, SIZE> &ptr)
{
	// Push data from bin to a queue of bin parts
	if(ptr.bins)
		for(uint32 i = 0; i < ptr.n_bins; ++i)
			if(ptr.bins[i])
				ptr.PushBinToQueue(i, true);
}


//************************************************************************************************************
// CWSplitter class - wrapper for multithreading purposes
//************************************************************************************************************

//----------------------------------------------------------------------------------
template <typename KMER_T, unsigned SIZE> class CWSplitter {
	CPartQueue *pq;
	CBinPartQueue *bpq;
	CMemoryPool *pmm_fastq;

	CSplitter<KMER_T, SIZE> *spl;
	uint64 n_reads;

public:
	CWSplitter(CKMCParams &Params, CKMCQueues &Queues);
	~CWSplitter();

	void operator()();
	void GetTotal(uint64 &_n_reads);
};

//----------------------------------------------------------------------------------
// Constructor
template <typename KMER_T, unsigned SIZE> CWSplitter<KMER_T, SIZE>::CWSplitter(CKMCParams &Params, CKMCQueues &Queues)
{
	pq		  = Queues.part_queue;
	bpq		  = Queues.bpq;
	pmm_fastq = Queues.pmm_fastq;

	spl = new CSplitter<KMER_T, SIZE>(Params, Queues);
}

//----------------------------------------------------------------------------------
// Destructor
template <typename KMER_T, unsigned SIZE> CWSplitter<KMER_T, SIZE>::~CWSplitter()
{
}

//----------------------------------------------------------------------------------
// Execution
template <typename KMER_T, unsigned SIZE> void CWSplitter<KMER_T, SIZE>::operator()()
{
	// Splitting parts
	while(!pq->completed())
	{
		uchar *part;
		uint64 size;
		if(pq->pop(part, size))
		{
			spl->ProcessReads(part, size);
			pmm_fastq->free(part);
		}
	}
	spl->Complete();
	bpq->mark_completed();

	spl->GetTotal(n_reads);

	delete spl;
	spl = NULL;
}

//----------------------------------------------------------------------------------
// Return statistics
template <typename KMER_T, unsigned SIZE> void CWSplitter<KMER_T, SIZE>::GetTotal(uint64 &_n_reads)
{
	if(spl)
		spl->GetTotal(n_reads);

	_n_reads = n_reads;
}

#endif

// ***** EOF
